#pragma once

#define MARSGAMING_MM4_NUM_PROFILES 5
#define MARSGAMING_MM4_NUM_RESOLUTIONS_PER_PROFILE 5
#define MARSGAMING_MM4_NUM_BUTTONS 19
#define MARSGAMING_MM4_NUM_LED 1
#define MARSGAMING_MM4_RES_MIN 50
#define MARSGAMING_MM4_RES_MAX 16400
#define MARSGAMING_MM4_RES_SCALING 50
